
%    ---------------------------------------------------------------------
%
%    MATLAB/MEX Transforming image completion
%    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see http://www.gnu.org/licenses/.
%
%    ---------------------------------------------------------------------
 
function [T S U] = colour_nnfield(nnfield,transformation)
    
    [R C k] = size(nnfield);
    
    magmax = sqrt(R^2 + C^2);
    
    [cc,rr] = meshgrid(1:C,1:R);
   
    Rmat = nnfield(:,:,1) - rr;
    Cmat = nnfield(:,:,2) - cc;
    
    % 0<=theta<=1
    thetamat = (atan2(Cmat,Rmat)/(2*pi))+0.5;
    
    % 0<=mag<=1
    magmat = sqrt(Rmat.^2 + Cmat.^2) / magmax;
    
    hsvim = 0.6*ones(R,C,3);
    hsvim(:,:,1) = thetamat;
    hsvim(:,:,2) = magmat;
    
    T = uint8(255*hsv2rgb(hsvim));
    
    if( nargout>1 )
        S = 0;
        
        if( transformation==1 || transformation==3 )
            % also visualize rotation
            
            S = uint8(255*repmat(nnfield(:,:,3)/8,[1 1 3]));
        elseif (transformation==2 )
            % also visualize scale
            
            S = uint8(255*repmat(nnfield(:,:,3),[1 1 3]));
        end
        
        if( nargout>2 )
            U = 0;
            
            if( transformation==3 )
                % also visualize scale
            
                U = uint8(255*repmat(nnfield(:,:,4),[1 1 3]));
            end
        end
    end
    
end