
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

#include "transformic_pmnn.h"
#include "mex.h"

template<class patch>
int transformic_patchMatch_db(double * const nnfield, images const x, int const iterations, int const iterations_before_previous, double const alpha, int const use_nnfield0)
{
    // construct patches object
    patch * patches;
    patches = new patch[x.TN];

    patch * ptr_patches = NULL;
    double * ptr_Tind = NULL;
    double * ptr_Sind = NULL;
    int r, c, tr, tc;

    patch p;
    double * p_parameters = new double[patch::num_parameters+1];

    /* -----------------------------------------------------------------
     * Initialization
     *
     * Initialize to best of copy NN or uniform random 
     * -----------------------------------------------------------------
     */
    
    // for each patch in T
    ptr_patches = patches;
    ptr_Tind = x.Tind;
    for( int i=0; i<x.TN; i++ )
    {
	tr = (int)*ptr_Tind;
	tc = (int)*(++ptr_Tind);
	ptr_Tind++;
	ptr_Tind++;

	ptr_patches->assign_location(tr,tc,i);

	ptr_patches->initialize(x);
	ptr_patches++;
    }

    for( int it=0;it<iterations;it++ )
    { 
	/* -----------------------------------------------------------------
	 * Propagation
	 * -----------------------------------------------------------------
	 */

	// for each patch in T
 	ptr_patches = patches;
	ptr_Tind = x.Tind;
	for( int i=0; i<x.TN; i++ )
	{
	    r = (int)*ptr_Tind;
	    c = (int)*(++ptr_Tind);
	    ptr_Tind++;
	    ptr_Tind++;

	    // define patch offset to propagate match from
	    int coffset = -1, roffset = -1;
	    if( (it%2)==0 )
	    {
		coffset = 1;
		roffset = 1;
	    }
	    
	    for( int dir=0; dir<=1; dir++ )
	    {
		tr = r+(roffset*dir);
		tc = c+(coffset*(1-dir));

		if( (tc>=x.patchdbT_offset) &&
		    (tc<(x.TC-x.patchdbT_offset)) &&
		    (tr>=x.patchdbT_offset) &&
		    (tr<(x.TR-x.patchdbT_offset)) ) 
		{
		    int tind = (int)x.Tindimage[index2(tr,tc,x.TR)];
		    if( tind>=0 ) // if patch valid in T
			ptr_patches->consider(patches[tind],x);
		}
		
	    }
	    ptr_patches++;
	}

	/* -----------------------------------------------------------------
	 * Random search
	 * -----------------------------------------------------------------
	 */

	// for each patch in T
	ptr_patches = patches;
	for( int i=0; i<x.TN; i++ )
	{
	    ptr_patches->random(x,alpha);
	    ptr_patches++;
	}

	/* -----------------------------------------------------------------
	 * Consider nnfield0 as proposal nn patches
	 * -----------------------------------------------------------------
	 */

	if ( (use_nnfield0==1) && (it==iterations_before_previous) )
	{
	    ptr_patches = patches;
	    ptr_Tind = x.Tind;
	    for( int i=0; i<x.TN; i++ )
	    {
		tr = (int)*ptr_Tind;
		tc = (int)*(++ptr_Tind);
		ptr_Tind++;
		ptr_Tind++;

		p.assign_location(tr,tc,i);

		for( int j=0; j<patch::num_parameters; j++ )
		{
		    // position index parameter excluded
		    if( j<2 )
			p_parameters[j] = x.nnfield0[index3(tr,tc,j,x.TR,x.TC)];
		    else
			p_parameters[j+1] = x.nnfield0[index3(tr,tc,j,x.TR,x.TC)];
		}
		p_parameters[2] = x.Sindimage[index2(p_parameters[0],p_parameters[1],x.SR)];
		p.assign_parameters(p_parameters);

		ptr_patches->consider(p,x);

		ptr_patches++;
	    }
	}

    } // next iteration

/*     // how many/\*  calls *\/ */
/* /\*     // for each patch in T *\/ */
/* /\*     ptr_patches = patches; *\/ */
/* /\*     ptr_Tind = x.Tind; *\/ */
/* /\*     for( int i=0; i<x.TN; i++ ) *\/ */
/* /\*     { *\/ */
/* /\* 	tr = (int)*ptr_Tind; *\/ */
/* /\* 	tc = (int)*(++ptr_Tind); *\/ */
/* /\* 	ptr_Tind++; *\/ */
/* /\* 	ptr_Tind++; *\/ */

/* /\* 	mexPrintf("pm_db tr %i tc %i calls %i\n",tr,tc,ptr_patches->calls); *\/ */

/* /\* 	ptr_patches++; *\/ */
/* /\*     } *\/ */

    // construct final nnfield

    // for each patch in T
    ptr_patches = patches;
    ptr_Tind = x.Tind;
    for( int i=0; i<x.TN; i++ )
    {
	tr = (int)*ptr_Tind;
	tc = (int)*(++ptr_Tind);
	ptr_Tind++;
	ptr_Tind++;

	ptr_patches->get_parameters(p_parameters);
	for( int j=0; j<patch::num_parameters; j++ )
	{
	    // this random access into the nnfield array is very slow!
	    // exclude position index parameter
	    if( j<2 )
		nnfield[index3(tr,tc,j,x.TR,x.TC)] = p_parameters[j];
	    else
		nnfield[index3(tr,tc,j,x.TR,x.TC)] = p_parameters[j+1];
	}
	nnfield[index3(tr,tc,patch::num_parameters,x.TR,x.TC)] = sqrt( ptr_patches->get_dist() / ((double)x.pN*3.0) );

	ptr_patches++;
    }

    // clean up memory
    delete [] p_parameters;
    delete [] patches;

    return 0;
}

// explicit instantiations to compile for
template int transformic_patchMatch_db<patch_tr_db>(double * const nnfield, images const x, int const iterations, int const iterations_before_previous, double const alpha, int const use_nnfield0);
template int transformic_patchMatch_db<patch_fl_db>(double * const nnfield, images const x, int const iterations, int const iterations_before_previous, double const alpha, int const use_nnfield0);
template int transformic_patchMatch_db<patch_il_db>(double * const nnfield, images const x, int const iterations, int const iterations_before_previous, double const alpha, int const use_nnfield0);
template int transformic_patchMatch_db<patch_rot_db>(double * const nnfield, images const x, int const iterations, int const iterations_before_previous, double const alpha, int const use_nnfield0);
template int transformic_patchMatch_db<patch_sc_db>(double * const nnfield, images const x, int const iterations, int const iterations_before_previous, double const alpha, int const use_nnfield0);
template int transformic_patchMatch_db<patch_sim_db>(double * const nnfield, images const x, int const iterations, int const iterations_before_previous, double const alpha, int const use_nnfield0);
template int transformic_patchMatch_db<patch_simb_db>(double * const nnfield, images const x, int const iterations, int const iterations_before_previous, double const alpha, int const use_nnfield0);
