
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

#include "transformic_pmnn.h"
#include <math.h>

/**************************************************************************
 * patch_il
 **************************************************************************/

void patch_il_db::get_patch(images const & x, double * Simagedb, double * patch) const 
{
    // patch should be allocated length x.pN * 3

    double * ptr_Patchind = x.patchind;
    double * const ptr_Sdb = &(Simagedb[index2(0,s,x.patchdbS_w*x.patchdbS_w*3)]);
    double * ptr_ptr_Sdb;
    int r, c;
    for( int j=0; j<x.pN; j++ )
    {
	r = (int)*ptr_Patchind;
	c = (int)*(++ptr_Patchind);
	ptr_Patchind++;

	ptr_ptr_Sdb = &(ptr_Sdb[index3(0,r,c,3,x.patchdbS_w)]);
	for(int k=0; k<3; k++) 
	{
	    *patch = std::min(std::max(0.0,*ptr_ptr_Sdb + illum),255.0);
	    patch++;
	    ptr_ptr_Sdb++;
	}
    }

    return;
}

void patch_il_db::assign_parameters(double * parameters) 
{
    // parameters should be allocated length patch_il::num_parameters+1
    sr = (int)parameters[0];
    sc = (int)parameters[1];
    s = (int)parameters[2];
    illum = parameters[3];
}

void patch_il_db::get_parameters(double * parameters) const 
{
    // parameters should be allocated length patch_il::num_parameters+1
    parameters[0] = (double)sr;
    parameters[1] = (double)sc;
    parameters[2] = (double)s;
    parameters[3] = illum;
}

void patch_il_db::random_parameters(images const & x) 
{
    // random illum
    illum = 510.0*((double)rand()/(double)RAND_MAX) - 255.0;
}

void patch_il_db::optimize_parameters(images const & x) {
    // optimize illum for current (sr,sc)

    illum = 0.0;
    int num_votes = 0;

    get_patch(x,x.Sdb,x.patchS);
    double * ptr_PatchS = x.patchS;
    double * ptr_PatchT = &(x.Tdb[index2(0,t,x.pN*3)]);
    double * ptr_PatchTmask = &(x.Tmaskdb[index2(0,t,x.pN)]);
    for( int j=0; j<x.pN; j++ )
    {
	// if Tmask does not indicate not to count the value of the ssd at this pixel
	if( !((*ptr_PatchTmask)>0.5) )
	{
	    for(int k=0; k<3; k++) 
	    {
// 		mexPrintf("%f-%f=%f ",*ptr_PatchT,*ptr_PatchS,*ptr_PatchT - *ptr_PatchS);

		num_votes++;
		illum += *ptr_PatchT - *ptr_PatchS;
		ptr_PatchS++;
		ptr_PatchT++;
	    }
	} else {
	    for(int k=0; k<3; k++) 
	    {
		ptr_PatchS++;
		ptr_PatchT++;
	    }
	}
	ptr_PatchTmask++;
    }

    if( num_votes>0 )
	illum /= (double)num_votes;

   //  mexPrintf("\nOpt illum tr %i tc %i sr %i sc %i num_votes %i %f\n",tr,tc,sr,sc,num_votes,illum);
    
}

void patch_il_db::random(images const & x, double const alpha) {
    // random

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s;
    double best_illum = illum;
    double best_dist = dist;

    // random search
    double search_radius = (double)std::max(x.TC,std::max(x.TR,std::max(x.SC,x.SR)));
    double search_radius_illum = 255.0;
    while( search_radius > 1.0 )
    {
	sr += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));
	sc += (int)(search_radius*((2.0*((double)rand()/(double)RAND_MAX))-1.0));

	// if this points to a valid patch
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);
		else
		{
		    //random_parameters(x);

		    // GENERALIZED PATCH MATCH
		    illum += search_radius_illum*((2.0*((double)rand()/(double)RAND_MAX))-1.0);
		}

		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    illum = best_illum;
		    dist = best_dist;
		} else {
		    // if have found a better match, update best
		    best_sr = sr;
		    best_sc = sc;
		    best_s = s;
		    best_illum = illum;
		    best_dist = dist;
		}	
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		illum = best_illum;
		dist = best_dist;
	    }						
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    illum = best_illum;
	    dist = best_dist;
	}

	search_radius *= alpha;
	search_radius_illum *= alpha;
    }
}

void patch_il_db::consider(patch_il_db const & p, images const & x) {
   // consider parameters based on parameters at another location

    // backup best parameters so far
    int best_sc = sc, best_sr = sr, best_s = s;
    double best_illum = illum;
    double best_dist = dist;

    int p_location[3];
    double p_parameters[patch_il_db::num_parameters+1];
    p.get_location(p_location);
    p.get_parameters(p_parameters);

    if( p_location != NULL && p_parameters != NULL )
    {
	// p_location and p_parameters assigned
	illum = p_parameters[3];
	sr = (int)(tr - p_location[0]) + (int)p_parameters[0];
	sc = (int)(tc - p_location[1]) + (int)p_parameters[1];

	// check transformed patch match from S is valid
	if( (sr>=x.patchdbS_offset)&&(sr<(x.SR-x.patchdbS_offset))&&(sc>=x.patchdbS_offset)&&(sc<(x.SC-x.patchdbS_offset)) )
	{
	    s = x.Sindimage[index2(sr,sc,x.SR)];

	    if( s>=0 )
	    {
		if( x.optimize>0 )
		    optimize_parameters(x);
		
		calculate_dist(x);

		if( dist > best_dist )
		{
		    // if have not found a better match, restore previous best parameters
		    sr = best_sr;
		    sc = best_sc;
		    s = best_s;
		    illum = best_illum;
		    dist = best_dist;
		}
	    } else {
		// if proposal not valid, restore previous best parameters
		sr = best_sr;
		sc = best_sc;
		s = best_s;
		illum = best_illum;
		dist = best_dist;	
	    }
	} else {
	    // if proposal not valid, restore previous best parameters
	    sr = best_sr;
	    sc = best_sc;
	    s = best_s;
	    illum = best_illum;
	    dist = best_dist;
	}
    } //else do nothing
}
