
/*    ---------------------------------------------------------------------
 *
 *    MATLAB/MEX Transforming image completion
 *    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 *    ---------------------------------------------------------------------
 */ 

#include "transformic_pmnn.h"
#include <map>

template<class patch>
void transformic_vote_weightedmedian_db(images const & x, double * const nnfield, std::list<std::pair<double,double> > * const votes, double * const Tweight_out)
{
    patch p;
	double * p_parameters = new double[patch::num_parameters+1];

    // for each patch in T
    for(int c=x.patch_offset; c<(x.TC-x.patch_offset); c++)
    {
		for(int r=x.patch_offset; r<(x.TR-x.patch_offset); r++)
		{
			if( x.Tdomain[index2(r,c,x.TR)]>0.0 )
			{
				int sr = (int)nnfield[index3(r,c,0,x.TR,x.TC)];
				int sc = (int)nnfield[index3(r,c,1,x.TR,x.TC)];
				int s = x.Sindimage[index2(sr,sc,x.SR)];

				if( s>=0 )
				{
					p_parameters[0] = (double)sr;
					p_parameters[1] = (double)sc;
					p_parameters[2] = (double)s;
					for( int i=2; i<patch::num_parameters; i++ )
					{
						p_parameters[i+1] = nnfield[index3(r,c,i,x.TR,x.TC)];
					}
					p.assign_parameters(p_parameters);
		
					p.get_patch(x,x.Sdb,x.patchS);

					if( x.patchS == NULL )
					{
						//handle
					} else {
						// if votes from this patch have some weight
						if( (x.Tweight[index2(r,c,x.TR)]>0) )
						{
							double * ptr_PatchS = x.patchS;
							double * ptr_Patchind = x.patchind;
							for( int j=0; j<x.pN; j++ )
							{
								int tr = r+(int)*ptr_Patchind-x.patchdbS_offset;
								int tc = c+(int)*(++ptr_Patchind)-x.patchdbS_offset;
								double tweight = x.Tweight[index2(r,c,x.TR)];
								for( int k=0; k<3; k++ )
								{
									votes[index3(tr,tc,k,x.TR,x.TC)].push_back(std::pair<double,double>(*ptr_PatchS,tweight));
									ptr_PatchS++;
								}

								Tweight_out[index2(tr,tc,x.TR)] += tweight;

								ptr_Patchind++;
							}

						}
					}
				}
			}
		}
	}
	delete [] p_parameters;
}
