
%    ---------------------------------------------------------------------
%
%    MATLAB/MEX Transforming image completion
%    Copyright (C) 2011 Alex Mansfield (alex.mansfield@vision.ee.ethz.ch)
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see http://www.gnu.org/licenses/.
%
%    ---------------------------------------------------------------------

function visualize_nnfield_click(obj,events,im_h,T,S,nnfield,patch_w,transformation,patchdb,patchind,Sdb,Sindimage)
    
    % get point clicked on
    fig_h = imgca;
    [nrows,ncols] = size(get(im_h,'CData'));
    xdata = get(im_h,'XData');
    ydata = get(im_h,'YData');
    point = get(fig_h,'CurrentPoint');
    point_im = point(1,1:2);
    tc = ceil(ceil(axes2pix(ncols,xdata,point_im(1)))/3);
    tr = ceil(axes2pix(nrows,ydata,point_im(2)));
    
    sr = nnfield(tr,tc,1)+1;
    sc = nnfield(tr,tc,2)+1;
    
    [TR TC colours] = size(T);
    patch_offset = (patch_w-1)/2;
    
    % check point_im not too near the border
    if( (tr>=patch_offset)&&(tr<=(TR-patch_offset))&&(tc>=patch_offset)&&(tc<=(TC-patch_offset))&&...
            (sr>=patch_offset)&&(sr<=(TR-patch_offset))&&(sc>=patch_offset)&&(sc<=(TC-patch_offset)) )
        
        Tpatch = T(tr-patch_offset:tr+patch_offset,tc-patch_offset:tc+patch_offset,:);
        Spatch = transformic_get_patch(S,reshape(nnfield(tr,tc,:),[],1),patch_w,transformation,patchdb,patchind,Sdb,Sindimage);
        
        if( patchdb>0 )
            % this code is taken directly from transformic_inpaint
            patch_offset = (patch_w-1)/2; 
            patchdbT_w = patch_w;
            patchdbT_offset = patch_offset;
            patchdbS_w = patchdbT_w+2;
            patchdbS_offset = patchdbT_offset+1;
            if( transformation==2 || transformation==3 || (transformation==6) ) %if scaling
                patchdbS_w = patchdbT_w*2-1;
                patchdbS_offset = patchdbT_offset*2;
            end
            
            if( (transformation==0) || (transformation==4) || (transformation==2) || (transformation==5) )
                % square patches
                [I J] = ind2sub([patch_w patch_w],(1:(patch_w*patch_w)));
                patchind = [I;J] - 1 + patchdbS_offset-patchdbT_offset;
            elseif( transformation==1 || transformation==3 || (transformation==6) )
                % circle patches
                
                % indices into all points
                [nn,mm] = meshgrid(1:patchdbT_w,1:patchdbT_w);
                % normalize them
                nn = nn - patchdbT_offset - 1;
                mm = mm - patchdbT_offset - 1;
                
                discmask = ((mm.^2 + nn.^2) <= patchdbT_offset^2);
                
                [I J] = ind2sub([patchdbT_w patchdbT_w],find(discmask>0)');
                patchind = [I;J] - 1 + patchdbS_offset-patchdbT_offset;
            end
            
            % reshape Spatch according to indices
            Spatchvec = Spatch;
            Spatch = zeros(patchdbT_w,patchdbT_w,colours);
            for i = 1:size(patchind,2)
                Spatch(patchind(1,i)+1-patchdbS_offset+patchdbT_offset,patchind(2,i)+1-patchdbS_offset+patchdbT_offset,1) = Spatchvec(i,1);
                Spatch(patchind(1,i)+1-patchdbS_offset+patchdbT_offset,patchind(2,i)+1-patchdbS_offset+patchdbT_offset,2) = Spatchvec(i,2);
                Spatch(patchind(1,i)+1-patchdbS_offset+patchdbT_offset,patchind(2,i)+1-patchdbS_offset+patchdbT_offset,3) = Spatchvec(i,3);
            end
        end
        
        % show source position
        figure(1);
        Tpos = T;
        Tpos(sr-patch_offset:sr+patch_offset,sc-patch_offset:sc+patch_offset,:) = 255 - Tpos(sr-patch_offset:sr+patch_offset,sc-patch_offset:sc+patch_offset,:);
        set(im_h,'CData',Tpos/255);
        
        % show patches
        figure(2);
        subplot(1,3,1);
        imshow(uint8(Tpatch));
        title(['Tpatch (tr ' int2str(tr) ' tc ' int2str(tc) ')']);
        subplot(1,3,2);
        imshow(uint8(T(sr-patch_offset:sr+patch_offset,sc-patch_offset:sc+patch_offset,:)));
        title(['Spatch (untransformed)']);
        subplot(1,3,3);
        imshow(uint8(Spatch));
        title(['Spatch (' num2str(nnfield(tr,tc,:)) ')']);
        
        Tpatch(:,:,1)
        Spatch(:,:,1)
        Tpatch(:,:,1)-Spatch(:,:,1)
        
        sqrt(sum(sum(sum((Tpatch-Spatch).^2)))/ (3*patch_w*patch_w))
        
    end

end